#ifndef residuals_calc_h
#define residuals_calc_h 1

#include "marlin/Processor.h"

#include "lcio.h"

#include <string>
#include <vector>
#include <map>
#include <cmath>
#include <algorithm>
#include <sstream>
#include <iostream>
#include <climits>
#include <cfloat>
#include "sys/types.h"
#include "sys/sysinfo.h"

#include <gsl/gsl_rng.h>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <EVENT/LCStrVec.h>
#include <EVENT/LCRelation.h>
#include <EVENT/SimTrackerHit.h>
#include <EVENT/CalorimeterHit.h>
#include <IMPL/CalorimeterHitImpl.h>
//#include <EVENT/SimCalorimeterHit.h>
#include <EVENT/Cluster.h> // for CalorimeterHitVec

#include <IMPL/TrackerHitPlaneImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/TrackImpl.h>

#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCRelationImpl.h>

#include <UTIL/CellIDDecoder.h>
#include <UTIL/CellIDEncoder.h>
#include <UTIL/LCIterator.h>
#include <UTIL/ILDConf.h>
#include <UTIL/BitSet32.h>
#include <UTIL/LCRelationNavigator.h>

#include "DDRec/Surface.h"
#include "DD4hep/LCDD.h"
#include "DD4hep/DD4hepUnits.h"
#include "DDRec/SurfaceManager.h"
#include "DDRec/API/IDDecoder.h"

#include "marlin/Global.h"
#include "marlin/ProcessorEventSeeder.h"
#include "marlin/VerbosityLevels.h"
#include <marlinutil/HelixClass.h>
#include "MarlinTrk/MarlinDDKalTestTrack.h"
#include "MarlinTrk/IMarlinTrkSystem.h"
#include "MarlinTrk/MarlinTrkUtils.h"
#include "MarlinTrk/HelixTrack.h"
#include "MarlinTrk/HelixFit.h"
#include "MarlinTrk/IMarlinTrack.h"
#include "MarlinTrk/Factory.h"
#include "MarlinTrk/MarlinTrkDiagnostics.h"
//#include "MarlinTrk/TrackFitter.h"

#include <AIDA/AIDA.h>
#include <AIDA/IAnalysisFactory.h>
#include <AIDA/IHistogramFactory.h>

#include <TH1F.h>
#include <TLorentzVector.h>
#include "TTree.h"
#include <TCanvas.h>
#include <TGraphAsymmErrors.h>
#include "TLegend.h"
class TTree;

using namespace lcio ;
using namespace marlin ;
using namespace AIDA ;

class ResidualsCalculator : public Processor {
		
public:
	
	virtual Processor*  newProcessor() { return new ResidualsCalculator; }
	
	ResidualsCalculator() ;
	
	// Initialisation - run at the beginning to start histograms, etc.
	virtual void init() ;
	
	// Called at the beginning of every run
	virtual void processRunHeader( LCRunHeader* run ) ;
	
	// Run over each event - the main algorithm
	virtual void processEvent( LCEvent * evt ) ;
	
	// Run at the end of each event
	virtual void check( LCEvent * evt ) ;
	
	// Called at the very end for cleanup, histogram saving, etc.
	virtual void end() ;
	
	// Call to get collections
	void getCollection(LCCollection*&, std::string, LCEvent*);

	// Get the subdetector ID from a collection or hit
	int getSubdetector(LCCollection* , UTIL::BitField64&);
	int getSubdetectorRel( LCCollection*, UTIL::BitField64&);
	
	int getSubdetector(TrackerHit*, UTIL::BitField64&);

	bool isReconstructable(MCParticle*& particle, std::string cut);

	void clearTreeVar();

	
protected:
	
	// Collection names for (in/out)put
	std::vector<std::string> m_inputTrackerHitCollections ;
	std::vector<std::string> m_inputTrackerHitRelationCollections ;
	std::string m_inputParticleCollection ;
	std::string m_inputTrackCollection ;
	std::string m_inputTrackRelationCollection;
	std::string m_notRecoMCColName;
	  //std::string m_inputTrackRelationCollection ;
	// Run and event counters
	int m_eventNumber ;
	int m_runNumber ;
	std::string _outFileName;
	std::string  _outFileType;
	// Track fit factory
	MarlinTrk::IMarlinTrkSystem* trackFactory;
	MarlinTrk::IMarlinTrkSystem* _trksystem;
	// Parameters
	double m_purity;
	double m_magneticField;
	std::map<std::string,double> m_particles;
	std::map<std::string,double> m_reconstructedParticles;
	
	int m_vertexBarrelID;
	std::string m_cuts;
	std::map<MCParticle*, std::vector<TrackerHit*> > particleHits;
	bool m_morePlots;

	// Plots 
	TCanvas *eff_vs_theta;
	TGraphAsymmErrors *g_eff_vs_theta;
	TH1F *h_theta_reconstructed ;
	TH1F *h_theta_reconstructable ;

	TCanvas *eff_vs_pt;
	TGraphAsymmErrors *g_eff_vs_pt;
	TH1F *h_pt_reconstructed ;
	TH1F *h_pt_reconstructable ;

	TH1F *h_pt_truth ;
	TH1F *h_pt_reco ;
	TH1F *h_pt_res ;
	TH1F *h_energy ;
	
	
	TH1I *h_particles ;
	 double ommcp,omtrack,truePt,recoPt,resPt,theta_mcp,pt_mcp, energy; 
	int particles_distribution;
	
	
	TTree *mctree ;
	std::vector<int > m_mcCat;
	std::vector<double > m_mcTheta;
	std::vector<double > m_mcPt;
	std::vector<int > m_mcIsDecayedInTracker;
	//std::vector<std::vector<int > > m_mcNHits;
	//std::vector<int > m_mcNHits_helper;
	std::vector<int > m_mcNHitsTot;
	std::vector<int > m_mcNHitsVXD;
	std::vector<int > m_mcNTracks; 
	std::vector<int > m_mcNTrkHits; 
	std::vector<int > m_mcThetaTrk; 
	std::vector<int > m_mcPtTrk; 
	std::vector<int > m_mcPhiTrk; 
	std::vector<int > m_mcNTracksCone; 
};

#endif // residuals_calc_h



